// #define borracursor

/******************************************************************
* Reconstruccion algebraica de volumenes usando el metodo SART
* ------------------------------------------------------------
* Programa para reconstruir volumenes a partir de una coleccion de
* proyecciones suyas. 
*
*
* TAG (250 de 64x64x64), lambda 0.06
*            tiempo    error   speedup
*   Xmipp   45.207s  0.003764   1
*   sart -v  1.140s  0.001913   39.7
*   sart -b  1.016s  0.002113   44.5
*
* TAG (1000 de 128x128x128)
*   Xmipp   ~1437s (24min)
*   sart -v  53.887s            26.7
*   sart -b  33.390s            43.0
*
* churrins (600 de 64x64x64)
*   Xmipp    106.751s           
*   sart -v    2.788s           38.3
*   sart -b    2.476s           43.1
*
* DNA (6995 de 81x81x81), todas las reconstrucciones son cutres
*   Xmipp   ~2472s (41min)            
*   sart -v   54.723s           45.1
*   sart -b   49.111s           50.3
*
*
* Proyecto de Fin de Carrera, Universidad Complutense de Madrid
* Facultad de Informatica, 2006/2007
*
* David Garcia Soriano
* Enrique Martin Martin
* David Romero Laorden
******************************************************************/

#include <XmippData/xmippImages.hh>
#include <XmippData/xmippSelFiles.hh>
#include <XmippData/xmippMatrices3D.hh>
#include <XmippData/xmippMatrices2D.hh>
#include <XmippData/xmippMatrices1D.hh>
#include <XmippData/xmippGeometry.hh>
#include <XmippData/xmippVolumes.hh>
#include <XmippData/xmippProjection.hh>
#include <iostream>
#include <cassert>
#include <cstring>
#include <cstdlib>
#include <algorithm>
#include <getopt.h>
#include <cmath>
#include <ctime>
using namespace std;

#ifdef borracursor
#include <curses.h>
#endif

#include "cohen3d.cpp"
#include "sart.hh"

/** Muestra las opciones de llamada del programa
*/
void usage ( char* ruta, int exit_code = 1){
    cout << "Usage: " << ruta << " options {-b ^ -v}" << endl;
    cout << "   -? --help              display this help and exit" << endl;
    cout << "   -i --input  filename   input selfile [img.sel]" << endl;
    cout << "   -o --output filename   output volume [flipas.vol]" << endl;
    cout << "   -l --lambda l          relaxation factor [0.01] (recommended range 0.0 - 0.1)" << endl;
    cout << "   -b --blob              use Kaisser-Bessel basis functions" << endl;
    cout << "   -v --voxel             use voxel basis functions" << endl;
    cout << "   -n --noit              number of iterations through all projections [1]" << endl;
    cout << "   -s --sort   n          sort projections so as to maximize minimum angle with last n images (0 = random)" << endl;
    cout << endl;
    exit (exit_code);
}

// Libera los recursos correspondientes, y termina el programa con un mensaje de error si ret != 0
void acaba(int ret) {
#ifdef borracursor  
  curs_set(1);
  endwin();
#endif  
  if (ret) exit(ret);
}


nreal volumen[maxtam][maxtam][maxtam];

/* Rayos que intersecan a un voxel y su peso */
vector<pair<pair<int, int>, nreal> > rayosinters[maxtam][maxtam][maxtam];

/// array global con los factores de normalizacion por cada pixel (i,j)
nreal normFactor[maxtam][maxtam];
nreal proyeccion[maxtam][maxtam];  // imagen proyectada

inline int signo(nreal x)                 { return x > 0 ? 1 : x < 0 ? - 1 : 0; }
inline nreal sigmedio(nreal x, int signo) { return signo > 0 ? ceil(x - 0.5 + eps) + 0.5 : floor(x - 0.5 - eps) + 0.5; }

// Estamos suponiendo que el punto p est�dentro del volumen o cerca de �, es esto cierto?
int siguerayo(int Zdim, int Ydim, int Xdim, punto p, punto v, coordvoxel* res) { // v est�normalizado
  int tam = 0;
  if (fabs(v.x) < eps) v.x = eps;
  if (fabs(v.y) < eps) v.y = eps;
  if (fabs(v.z) < eps) v.z = eps;
  
  // Si p est�dentro del volumen, lo saca
  nreal t = Xdim + Ydim + Zdim; // al estar normalizado, con esto se sale del volumen
  if (p.x >= -0.5 && p.x <= Xdim + 0.5 && p.y >= 0.5 && p.y <= Ydim + 0.5 && p.z >= -0.5 && p.z <= Zdim + 0.5)
    p -= t * v;
    
  // Pilla un punto suficientemente externo
  t *= 4;
  punto q(p + t * v);
  if (!clip(p.x, p.y, p.z, q.x, q.y, q.z, -0.5, -0.5, -0.5, Xdim + 0.5, Ydim + 0.5, Zdim + 0.5)) return 0;
  
  int sx = signo(v.x), sy = signo(v.y), sz = signo(v.z);
  nreal trestante = sqrt(p.dist(q) + eps);
  for (trestante = p.dist(q) + eps; trestante > 0; trestante -= t) {
    // Calcula el t de salida del voxel
    t   = (sigmedio(p.x, sx) - p.x) / v.x;
    t = min( t, (sigmedio(p.y, sy) - p.y) / v.y);
    //t <?= (sigmedio(p.y, sy) - p.y) / v.y;
    t = min( t, (sigmedio(p.z, sz) - p.z) / v.z);
    //t <?= (sigmedio(p.z, sz) - p.z) / v.z;
    
    // Calcula las coordenadas del voxel
    unsigned x = (unsigned)round(p.x + t / 2 * v.x), y = (unsigned)round(p.y + t / 2 * v.y), z = (unsigned)round(p.z + t / 2 * v.z);
    if (x < Xdim && y < Ydim && z < Zdim) res[tam++] = coordvoxel(x, y, z, t);
    p += t * v;
  }
//  assert(tam <= Xdim + Ydim + Zdim - 2);
  return tam;
}



/** Calcula y devuelve la proyeccion del volumen sobre un plano determinado por la matriz de rotacion.
* Tambien asigna el array global de 'peso' con los pesos de cada voxel a cada pixel y el array global
* normFactor con el factor de normalizacion por cada pixel
*/
void proyectarayos( 
  int Zdim, int Ydim, int Xdim,         /// Dimensiones de la proyeccion 
  const matrix2D<double>  &eulerMatrix, /// Matriz de Euler de la rotacion 
  punto direct                          /// Vector director de los rayos de la proyeccion
){  
  int i, j, k, midX = Xdim / 2, midY = Ydim / 2, midZ = Zdim / 2;
  for (i = 0; i < Zdim; ++i) for (j = 0; j < Ydim; ++j) for (k = 0; k < Xdim; ++k) rayosinters[i][j][k].clear();
  for( i = 0; i < Ydim; i++ )
    for( j = 0; j < Xdim; j++ ) {
      /* Calcula un punto por el que pasa el rayo */
      matrix1D<double>  puntoVoxel(3);  /// Punto del espacio donde esta el voxel
      matrix1D<double>  puntoPixel(3);  /// Punto del espacio donde esta el pixel
      puntoPixel(0) = j - midX;
      puntoPixel(1) = i - midY;      
      puntoPixel(2) = 0;  // esto la verdad es que me la pela que valor tenga
      puntoPixel = eulerMatrix * puntoPixel;     
      puntoPixel(0) += midX; puntoPixel(1) += midY; puntoPixel(2) += midZ;

      // Calcula los voxeles que interseca (como mucho Xdim + Ydim + Zdim - 2)
      coordvoxel voxinters[maxtam * 3];
      int tam = siguerayo(Zdim, Ydim, Xdim, (punto)puntoPixel, direct, voxinters);
      assert(tam <= Zdim + Ydim + Xdim);
      proyeccion[i][j] = 0.0;
      normFactor[i][j] = 0.0;
      for (k = 0; k < tam; ++k) {
        coordvoxel p(voxinters[k]);
        normFactor[i][j] += p.peso;
        proyeccion[i][j] += volumen[p.z][p.y][p.x] * p.peso;

        // A�de a rayosinters ((i, j), peso)
        rayosinters[p.z][p.y][p.x].push_back(make_pair(make_pair(i, j), p.peso));
      }
      if (!tam) normFactor[i][j] = 1.0; // evita divisi� por cero
    }
} 

void proyectavoxels(int Zdim, int Ydim, int Xdim,
const matrix2D<double>  &eulerMatrix, /// Matriz de Euler de la rotacion (la traspuesta de la de proyectarayos!)
punto direct                          /// Vector director de los rayos de la proyeccion
) {
  const int midZ = (Zdim + 1) / 2, midY = (Ydim + 1) / 2, midX = (Xdim + 1) / 2;
  memset(normFactor, 0, sizeof(normFactor));
  memset(proyeccion, 0, sizeof(proyeccion));
  for (int z = 0; z < Zdim; ++z) for (int y = 0; y < Ydim; ++y) for (int x = 0; x < Xdim; ++x) {
    // Calcula la proyeccin del centro del voxel
    matrix1D<double> punto(3);
    punto(0) = x - midX;
    punto(1) = y - midY;
    punto(2) = z - midZ;
    punto = eulerMatrix * punto;
    double projy = punto(1) + midY, projx = punto(0) + midX;
    rayosinters[z][y][x].clear();

    // El centro del voxel cae en (projy, projx), que est en el pixel (pixely, pixelx)
    int pixely = (int)round(projy), pixelx = (int)round(projx), 
        dira = pixely <= projy ? -1 : 1, dirb = pixelx <= projx ? -1 : 1;
    double dify = projy - pixely, difx = projx - pixelx; 
    for (int a = 0, na = 0; na < 2; a += dira, ++na)
      for (int b = 0, nb = 0; nb < 2; b += dirb, ++nb) {
        double dist2 = (dify + a) * (dify + a) + (difx + b) * (difx + b);
        if (dist2 <= 0.75) {
          int ny = pixely + a, nx = pixelx + b;              
          if (ny >= 0 && ny < Ydim && nx >= 0 && nx < Xdim) {
            double peso = 0.75 - dist2;
            rayosinters[z][y][x].push_back(make_pair(make_pair(ny, nx), peso));

            // Estas dos cosas son las que se hacen en el paso final
            normFactor[ny][nx] += peso;
            proyeccion[ny][nx] += volumen[z][y][x] * peso;
          }
        }
      }
  }
}

// Se supone que al llamar aqui proyeccion[] contiene ya las diferencias normalizadas
void backproyecta(int Zdim, int Ydim, int Xdim, nreal lambda = 1) {
  int x, y, z, i;
  for (z = 0; z < Zdim; ++z) 
    for (y = 0; y < Ydim; ++y)
      for (x = 0; x < Xdim; ++x) {
        int n = rayosinters[z][y][x].size();
//        printf("%i\n", n);
        if (!n) continue;
        
        nreal correccion = 0, sumawi = 0;
        for (i = 0; i < n; ++i) {
          pair<int, int> p = rayosinters[z][y][x][i].first;
          nreal peso = rayosinters[z][y][x][i].second,
                 diferencia = proyeccion[p.first][p.second];
          sumawi += peso;
          correccion += diferencia * peso;
        }
        volumen[z][y][x] += correccion / sumawi * lambda;
      }
}

int main( int argc, char** argv){
  FileName    fnProyAct,  /// Nombre del archivo de la proyeccion actual 
              fnVolume,   /// Nombre del archivo del volumen reconstruido 
              fnSelFile;  /// Nombre del archivo de seleccion de proyecciones
  bool        initialized;/// Si se ha inicializado el volumen
  Projection  proyAct;    /// Proyeccion actual (leida de un archivo)
  int         Xdim, 
              Ydim, 
              Zdim;       /// Dimensiones del volumen reconstruido
  int x, y, z;

  // Procesa las opciones de linea de comandos
  int hbproy = -1, hbback = -1, numiter = 1;
  float lambda;
  bool sort = false;

  const char* const shortopt = "i:o:l:n:bvs:?";
  const struct option longopt[] = {
    { "input",   1, NULL, 'i' },
    { "output",  1, NULL, 'o' },
    { "lambda",  1, NULL, 'l' },
    { "blob",    0, NULL, 'b' },
    { "voxel",   0, NULL, 'v' },
    { "noit",    1, NULL, 'n' },
    { "sort",    1, NULL, 's' },
    { "help",    0, NULL, '?' }
  };

  fnVolume = "flipas.vol";
  fnSelFile = "img.sel";
  lambda = 0.01;

  int tipobasis = 0, numsort;
  int next_option;
  do{
    next_option = getopt_long(argc, argv, shortopt, longopt, NULL );
    switch( next_option ){
      case 'i': fnSelFile = optarg; break;
      case 'o': fnVolume  = optarg; break;
      case 'l': lambda    = atof(optarg); break;
      case 'b': tipobasis |= 1; break;
      case 'v': tipobasis |= 2; break;
      case 'n': numiter = atoi(optarg); break;
      case 's': sort = true; numsort = atoi(optarg); break;
      case '?': usage(argv[0], 1 ); break;
      case -1 : break;
    }
  }while( next_option != -1 );
  if (optind != argc) usage(argv[0], 1);

  // Parametros por defecto
  if (tipobasis == 0 || tipobasis == 3) {
    puts("You must specify exactly one of -v and -b\n");
    usage(argv[0], 1);
  }
  bool raydriven = tipobasis == 2;



#ifdef borracursor  
  // Oculta el cursor
  initscr();
  curs_set(0);
#endif  


  // Lee el archivo de seleccion
  int numimg = 0;
  vector<string> nombreimg;
  vector<double> dirx, diry, dirz;
  SelFile     sel;        /// Archivo de seleccion de proyecciones 
  try{
    sel.read( fnSelFile );
  } catch (Xmipp_error XE) { acaba(0); cout << XE; exit(1);}
  srand(time(0));

  numimg = sel.ImgNo(); // este nmero cuenta las que no estn activas, pero el next() las da igual
  printf("%i images\n", numimg);
#ifdef opsort  
  if (sort) {
    numimg = 0;
    while (!sel.eof()) {
      printf("\rReading angle information... %5i", numimg);
      fflush(stdout);
      string s = sel.NextImg();
      nombreimg.push_back(s);
      ++numimg;

      // Si hay que ordenar, saca los angulos de cada proyeccion
      try { 
        proyAct.read(s, true);
        matrix2D<double> eulerMatrix(3, 3);
        matrix1D<double> direct(3);
        eulerMatrix = proyAct.eulert;
        eulerMatrix.getCol(2, direct);
        dirx.push_back(direct(0));
        diry.push_back(direct(1));
        dirz.push_back(direct(2));
      } catch (Xmipp_error XE) { acaba(0); cout << XE; exit(1);}
    }
  }
  if (sort) printf("\n\rSorting...\n\r");
  fflush(stdout);
#endif  

  // Ordena
  int indice[numimg], restantes[numimg], quedan = numimg;
  sort &= numimg > 2;
#ifdef opsort  
  if (sort) {
    int i, j, k;
    for (i = 0; i < numimg; ++i) restantes[i] = i;
    random_shuffle(&restantes[0], &restantes[numimg]);

    // Elige la primera al azar
    indice[0] = restantes[--quedan];

    for (i = 1; i < numimg; ++i) {
      int mejor;
      double res = 2;
      for (j = 0; j < quedan; ++j) {
        // Calcula el mayor valor absoluto del producto escalar con las dos anteriores
        double t = -2;
        for (k = max(0, i - numsort); k < i; ++k) {
          double presc = dirx[restantes[j]] * dirx[indice[k]] + 
                         diry[restantes[j]] * diry[indice[k]] + 
                         dirz[restantes[j]] * dirz[indice[k]];
          t = max(t, abs(presc));
        }

        if (t < res) {
          mejor = j;
          res = t;
        }
      }

      swap(restantes[mejor], restantes[--quedan]);
      indice[i] = restantes[quedan];
    }
  } else 
#endif    
  {
    nombreimg.resize(numimg);
    for (int i = 0; i < numimg; ++i) indice[i] = i;
  }

  // Hace el SART
  initialized = false;
  numiter = max(numiter, 1);
  printf("Running %i iteration(s) with lambda = %lf\n", numiter, lambda);
  for (int iteracion = 0; iteracion < numiter; ++iteracion) {
    for (int posimg = 0; posimg < numimg; ++posimg) {
      matrix2D<double> eulerMatrix(3,3); ///Matrix de Euler asociada a la rotacion
      matrix1D<double> direct(3);        ///Vector director de los rayos de la proyeccion

      if (sort)
        fnProyAct = nombreimg[indice[posimg]];
      else {
        nombreimg[posimg] = fnProyAct = sel.NextImg();
        if (posimg == numimg - 1) sort = true; // ya has leido todos los nombres
      }

      //Lee la proyeccion actual y la almacena en 'proyAct'
      try{ 
        bool apply_shifts = true;
        proyAct.read( fnProyAct, apply_shifts );
      }catch (Xmipp_error XE) { acaba(0); cout << XE; exit(1);}

      if( !initialized ){
        proyAct().get_dim( Ydim, Xdim );
        Zdim = max(Xdim, Ydim);
        printf("\rVolume dimensions (z, y, x): %i %i %i\n\r", Zdim, Ydim, Xdim);

        initialized = true;
      } 
      printf("\r  Iteration %3i     Image %5i     ", iteracion, posimg);
      cout << fnProyAct;
      fflush(stdout);

      eulerMatrix = proyAct.eulert;      // matriz de paso de proyeccion a universal
      eulerMatrix.getCol( 2, direct );   // direcci� del rayo (imagen del eje Z por eulerMatrix)

      if (raydriven)
        proyectarayos(Zdim, Ydim, Xdim, eulerMatrix, direct );
      else           
        proyectavoxels(Zdim, Ydim, Xdim, proyAct.euler, direct);

      // Calcula la diferencia
      for (int i = 0; i < Ydim; ++i)
        for (int j = 0; j < Xdim; ++j) {
          proyeccion[i][j] = (proyAct(i, j) - proyeccion[i][j]) / normFactor[i][j];
  //        printf("%f\n", proyeccion[i][j]);
        }
      backproyecta(Zdim, Ydim, Xdim, lambda);

        static int furcia = 0;
  //      if (++furcia == 1) break;
    }
    puts("\r");
  }
  acaba(0);

  //Escribe el volumen reconstruido
  VolumeXmipp vol(Zdim, Ydim, Xdim);
  int pos = 0;
  for (z = 0; z < Zdim; ++z) for (y = 0; y < Ydim; ++y) for (x = 0; x < Xdim; ++x, ++pos)
    vol(z, y, x) = volumen[z][y][x];
  vol.write( fnVolume );

  return 0;
}
