/* vim: set filetype=cpp : */

/* Funciones llamadas desde el programa principal:
 * inicializacuda(), finalizacuda(), llamagapu(...) */

/*
 * <= 512 threads / block
 * <= 65535 each dimension of thread blocks 
 * 16 multiprocesadores a 600 mhz
 * 768 mb
 * 16Kb shared memory / multiprocessor (block), divided into 16 banks
 * 32 threads / warp */
#include <cstdio>
#include <algorithm>
using namespace std;

#include "comun.h"
#include "myvector.h"

//#define DEBUG

#define pos2d(y,x) ((y) * Xdim + (x))
#define pos3d(z,y,x) ((z) * Ydim * Xdim + (y) * Xdim + (x))

#ifdef DEBUG
// Array para conseguir datos de lo que ocurre en la tarjeta
#define tamdebug 1000000
float *debug, debugcpu[tamdebug];
#endif

/* Coordenadas de un rayo y longitud del trozo de un voxel por el que pasa
 * (en realidad esto es lo mismo que lo anterior, aunque se usa para cosas distintas) */
struct inforayo {
  int pos;    // pos2d(y, x)
  float peso;
};

typedef myvector<inforayo> tiporayosporvoxel;
typedef float2 infoproyslice; // (normFactor, proyeccion)

/* Memoria total usada en GPU: aproximadamente 68n^3 + 4n^2 
 * ==> n < 227 */
namespace globalescuda {
  /* los que deberian ser 3D estan indexados como pos3d(z, y, x)
   * los 2D como pos2d(y, x) */
  tiporayosporvoxel *rayosporvoxel = 0;                   // myvector<inforayo>[Zdim][Ydim][Xdim]
  float *proyeccion = 0;                                  // float[Ydim][Xdim]
  float *volumen = 0;                                     // float[Ydim][Xdim][Xdim]
  float *volumenCPU = 0;                                  // float[Ydim][Xdim][Xdim]
  infoproyslice *slice = 0;                               // infoproyslice[Zdim][Ydim][Xdim]
  int Zdim, Ydim, Xdim;
};
using namespace globalescuda;



// Backproyectar (comun a ambas implementaciones)
// golpe es el numero de voxeles que cada kernel procesa (de un golpe).
__global__ void backproyecta(float *volumen, const float *proyeccion, const tiporayosporvoxel *rayosporvoxel, 
const int golpe, const int numvoxels, const float lambda, int hbbackproyecta
#ifdef DEBUG
, float *debug
#endif
) { 
  int pos = blockIdx.x * hbbackproyecta + threadIdx.x;
  pos *= golpe;
  for (int k = min(golpe, numvoxels - pos); k-- > 0; ++pos) {
    int n = rayosporvoxel[pos].size();
    if (n) { // si no es vacio no hace falta y asi ademas evita division por cero
      float correccion = 0, sumawi = 0;
      for (int i = 0; i < n; ++i) {  
        inforayo ri = rayosporvoxel[pos][i];
        // el acceso a proyeccion[ri.pos] es el 40% del tiempo de backproyeccion
        float diferencia = proyeccion[ri.pos], peso = ri.peso;
        sumawi += peso;
        correccion += diferencia * peso;
      } 
      volumen[pos] += correccion / sumawi * lambda;
    }
  }
}

// Reserva la memoria necesaria y guarda el puntero donde se copiara el volumen a la RAM del host
// Asume Zdim >= max(Xdim, Ydim)
void inicializacuda(float *res, int lz, int ly, int lx) {
  Zdim = lz; Ydim = ly; Xdim = lx;
  volumenCPU = res;

  int tam = sizeof(*proyeccion) * Ydim * Xdim;
  cudaMalloc((void **)&proyeccion, tam);

  // Puedo hacer el malloc del rayos inters porque nuestros vectors son de tamao fijo
  // Reservo algo mas de memoria de la necesaria porque paso de hacer
  // la comprobacion de fuera de volumen en borrarayosporvoxel
  int numvoxels = Zdim * Ydim * Xdim;
  tam = sizeof(*rayosporvoxel) * ((numvoxels + 511) >> 9) << 9; 
  cudaMalloc((void **)&rayosporvoxel, tam);
  
  // Inicializa el volumen a 0
  tam = sizeof(*volumen) * Zdim * Ydim * Xdim;
  cudaMalloc((void **)&volumen, tam);
  cudaMemset(volumen, 0, tam);

  cudaMalloc((void **)&slice, sizeof(*slice) * Zdim * Ydim * Xdim);

#ifdef DEBUG
  cudaMalloc((void **)&debug, sizeof(*debug) * tamdebug);
  cudaMemset(debug, 0xFE, sizeof(*debug) * tamdebug);
#endif  
}

// Libera memoria y copia el volumen calculado a la CPU
void finalizacuda() {
  if (volumenCPU)
    cudaMemcpy(volumenCPU, volumen, sizeof(*volumenCPU) * Zdim * Ydim * Xdim, cudaMemcpyDeviceToHost);
  cudaFree(volumen);
  cudaFree(rayosporvoxel);
  cudaFree(proyeccion);
  cudaFree(slice);
#ifdef DEBUG  
  cudaFree(debug);
#endif  
}

#include "voxelkernel.cu"
#include "raykernel.cu"


// Hace una iteracion de SART; eulerMatrix va de proyeccion a universal
int llamagpu(const float *proycpu, matrix3d eulerMatrix, float lambda, int hbproyecta, int hbbackproyecta, 
bool raydriven) {
  int ret = raydriven ?   raydrivenproj(proycpu, eulerMatrix, lambda, hbproyecta, hbbackproyecta)
                      : voxeldrivenproj(proycpu, eulerMatrix, lambda, hbproyecta, hbbackproyecta);
  if (ret) return ret;

  // Backproyecta
  {
    // Calcula golpe para asegurar que el numero de hilos es <= 65535
    dim3 dimBlock(hbbackproyecta, 1);
    int numvoxels = Zdim * Ydim * Xdim, h = (numvoxels + hbbackproyecta - 1) / hbbackproyecta,
        golpe = (h + 65534) / 65535;
    dim3 dimGrid(((numvoxels + golpe - 1) / golpe + hbbackproyecta - 1) / hbbackproyecta, 1);
    backproyecta<<< dimGrid, dimBlock >>>(volumen, proyeccion, rayosporvoxel, golpe, numvoxels, lambda, hbbackproyecta
#ifdef DEBUG
        , debug
#endif        
    );
    if (cudaGetLastError() != cudaSuccess) return ERROR_BACKPROY;
  }
  return 0;
}
