/* vim: set filetype=cpp : */
#include "punto.h"
#include "cohen3d.cpp"

// Devuelve -1, 0 o 1 segun el signo de un numero
__device__ inline int signo(nreal x) { return x > 0 ? 1 : x < 0 ? - 1 : 0; }

// Devuelve el siguiente multiplo de 0.5 >= x si signo > 0, y el anterior si signo < 0
__device__ inline nreal sigmedio(nreal x, int signo) 
{ return signo > 0 ? ceilf(x - 0.5f + eps) + 0.5f : floorf(x - 0.5f - eps) + 0.5f; }


/* Optimizaciones: se podra mezclar rayosporvoxel y volumen para minimizar accessos a memoria (por indvoxel) 
 * El 85% son los accesos a memoria */

// Proyeccion tiene la proyeccion del archivo; al acabar contendra la diferencia normalizada
// v es la direccion del rayo (normalizada)
__global__ void proyectarayos(const float *volumen, float *proyeccion, tiporayosporvoxel *rayosporvoxel,
const int Zdim, const int Ydim, const int Xdim, const matrix3d eulerMatrix, punto v, int posenbloque, int hbproyecta
#ifdef DEBUG
, float *debug
#endif
) {
  /* Calcula el correspondiente pixel; el hilo identifica el subcuadrado de 2x2 en la matriz
   * original y posenbloque uno de los 4 pixeles de ese subcuadrado. */
  const int midX = (Xdim + 1) / 2, midY = (Ydim + 1) / 2, midZ = (Zdim + 1) / 2;
  int pos = blockIdx.x * hbproyecta + threadIdx.x, i = pos / midX, j = pos % midX;
  i = (i << 1) | (posenbloque >> 1);
  j = (j << 1) | (posenbloque & 1);
  pos = pos2d(i, j);
  if (pos >= Ydim * Xdim || j >= Xdim) return; // por si el ultimo bloque o grid no esta completo

  // Calcula un punto del rayo (p) que no esta muy lejos del volumen
  punto p;
  int x = j - midX, y = i - midY;
  p.x = midX + eulerMatrix.val[0][0] * x + eulerMatrix.val[0][1] * y;
  p.y = midY + eulerMatrix.val[1][0] * x + eulerMatrix.val[1][1] * y;
  p.z = midZ + eulerMatrix.val[2][0] * x + eulerMatrix.val[2][1] * y;

  // Si p esta dentro del volumen, lo saca
  nreal t = Xdim + Ydim + Zdim; // al estar normalizado, con esto se sale del volumen
  if (p.x >= -0.5f && p.x <= Xdim + 0.5f && p.y >= 0.5f && p.y <= Ydim + 0.5f && p.z >= -0.5f && p.z <= Zdim + 0.5f) 
    p -= v * t;
    
  // Pilla un punto suficientemente externo
  t *= 4;
  punto q;
  q.assign(p + v * t);

  /* Clipea el segmento pq con el cubo de voxeles; si hay interseccion procesa el rayo.
   * Al hacer clip, el rayo cortara desde p hasta q */
  if ( clip(p.x, p.y, p.z, q.x, q.y, q.z, -0.5f, -0.5f, -0.5f, Xdim - 0.5f, Ydim - 0.5f, Zdim - 0.5f) )
  {
    int sx = signo(v.x), sy = signo(v.y), sz = signo(v.z);
    float normFactor = 0, proyteorica = 0;
    for (nreal trestante = p.dist(q); trestante > 0; trestante -= t) {
      // Calcula el t de salida del voxel (la longitud del trozo de rayo)
      t = fminf((sigmedio(p.x, sx) - p.x) / v.x, (sigmedio(p.y, sy) - p.y) / v.y);
      t = fminf(t, (sigmedio(p.z, sz) - p.z) / v.z);

      // Calcula las coordenadas del voxel
      int x = (int)roundf(p.x + t / 2 * v.x), y = (int)roundf(p.y + t / 2 * v.y), 
          z = (int)roundf(p.z + t / 2 * v.z);

      if (x >= 0 && x < Xdim && y >= 0 && y < Ydim && z >= 0 && z < Zdim) {
        int indvoxel = pos3d(z, y, x);
        inforayo ir;
        ir.pos = pos;   // (y, x) del rayo
        ir.peso = t;

        // manejar el array a mano (10%)
        tiporayosporvoxel *p = &rayosporvoxel[indvoxel];
        if (p->tam < tamvector) p->array[p->tam++] = ir;  // 50% del tiempo
//        rayosporvoxel[indvoxel].push_back(ir);
        
        normFactor += t;
        proyteorica += volumen[indvoxel] * t;
      }
      p += v * t;
    }

    // Calcula la diferencia normalizada
    proyeccion[pos] = (proyeccion[pos] - proyteorica) / normFactor;
  }
} 

__global__ void borrarayosporvoxel(tiporayosporvoxel *rayosporvoxel) {
  rayosporvoxel[blockIdx.x * 512 + threadIdx.x].tam = 0;
}

int raydrivenproj(const float *proycpu, matrix3d eulerMatrix, float lambda, int hbproyecta, int hbbackproyecta) {
  // Pilla la direccion del rayo (columna 2 de la matriz de Euler de 2D a 3D)
  punto direct;
  direct.x = eulerMatrix.val[0][2];
  direct.y = eulerMatrix.val[1][2];
  direct.z = eulerMatrix.val[2][2];

  // Evita divisiones por numeros cercanos a 0 (aunque cambie ligeramente el modulo de v)
  if (fabs(direct.x) < eps) direct.x = eps;
  if (fabs(direct.y) < eps) direct.y = eps;
  if (fabs(direct.z) < eps) direct.z = eps;

  // Copia la proyeccion medida (la del ImageXmipp)
  cudaMemcpy(proyeccion, proycpu, sizeof(*proyeccion) * Ydim * Xdim, cudaMemcpyHostToDevice);
  
  /* Borra los vectores de rayosporvoxel (solo hace falta poner el tamao a cero);
   * en modo emulacion, cambiar por el memset */
  borrarayosporvoxel<<< dim3((Zdim * Ydim * Xdim + 511) / 512, 1), dim3(512, 1)>>> (rayosporvoxel);
 
  // Proyecta (calcula los rayos que pasan por cada voxel)
  {
    dim3 dimGrid(((Ydim + 1) / 2 * ((Xdim + 1) / 2) + hbproyecta - 1) / hbproyecta, 1);
    dim3 dimBlock(hbproyecta, 1);
    for (int k = 0; k < 4; ++k)
      proyectarayos<<<dimGrid, dimBlock>>>(volumen, proyeccion, rayosporvoxel, Zdim, Ydim, Xdim, 
          eulerMatrix,
          direct,
          k, hbproyecta
#ifdef DEBUG          
, debug
#endif 
      );

    if (cudaGetLastError() != cudaSuccess) return ERROR_PROYRAYO;
  }

  return 0;
}
