// #define borracursor

/******************************************************************
* Reconstruccion algebraica de volumenes usando el metodo SART
* ------------------------------------------------------------
* Programa para reconstruir volumenes a partir de una coleccion de
* proyecciones suyas. 
*
*
* TAG (250 de 64x64x64), lambda 0.06
*            tiempo    error   speedup
*   Xmipp   45.207s  0.003764   1
*   sart -v  1.140s  0.001913   39.7
*   sart -b  1.016s  0.002113   44.5
*
* TAG (1000 de 128x128x128)
*   Xmipp   ~1437s (24min)
*   sart -v  53.887s            26.7
*   sart -b  33.390s            43.0
*
* churrins (600 de 64x64x64)
*   Xmipp    106.751s           
*   sart -v    2.788s           38.3
*   sart -b    2.476s           43.1
*
* DNA (6995 de 81x81x81), todas las reconstrucciones son cutres
*   Xmipp   ~2472s (41min)            
*   sart -v   54.723s           45.1
*   sart -b   49.111s           50.3
*
*
* Proyecto de Fin de Carrera, Universidad Complutense de Madrid
* Facultad de Informatica, 2006/2007
*
* David Garcia Soriano
* Enrique Martin Martin
* David Romero Laorden
******************************************************************/

#include <XmippData/xmippImages.hh>
#include <XmippData/xmippSelFiles.hh>
#include <XmippData/xmippMatrices3D.hh>
#include <XmippData/xmippMatrices2D.hh>
#include <XmippData/xmippMatrices1D.hh>
#include <XmippData/xmippGeometry.hh>
#include <XmippData/xmippVolumes.hh>
#include <XmippData/xmippProjection.hh>
#include <iostream>
#include <cassert>
#include <cstring>
#include <cstdlib>
#include <algorithm>
#include <getopt.h>
#include <cmath>
#include <ctime>
#include "comun.h"
using namespace std;

#ifdef borracursor
#include <curses.h>
#endif

float *volumen, *proyeccion;

/** Muestra las opciones de llamada del programa
*/
void usage ( char* ruta, int exit_code ){
    cout << "Usage: " << ruta << " options {-b ^ -v}" << endl;
    cout << "   -? --help     display this help and exit" << endl;
    cout << "   -i --input    input selfile filename [img.sel]" << endl;
    cout << "   -o --output   output volume filename [flipas.vol]" << endl;
    cout << "   -l --lambda   relaxation factor [0.01]" << endl;
    cout << "   -b --blob     use Kaisser-Bessel basis functions" << endl;
    cout << "   -v --voxel    use voxel basis functions" << endl;
    cout << "   -n --noit     number of iterations through all projections [1]" << endl;
    cout << "   -s --sort     sort projections before processing" << endl;
    cout << "      --tbproj   number of threads per block for projection" << endl;
    cout << "      --tbback   number of threads per block for backprojection" << endl;
    cout << endl;
    exit (exit_code);
}

// definidas en los *.cu
void inicializacuda(float *, int, int, int);
int llamagpu(const float *, matrix3d eulerMatrix, float lambda, int hbproy, int hbback, bool );
void finalizacuda();

// Libera los recursos correspondientes, y termina el programa con un mensaje de error si ret != 0
void acaba(int ret) {
#ifdef borracursor  
  curs_set(1);
  endwin();
#endif  
  switch (ret) {
    case ERROR_PROYVOXEL: puts("Voxel-driven projection kernel execution failed!!!"); break;
    case ERROR_PROYRAYO:  puts("Ray-driven projection kernel execution failed!!!"); break;
    case ERROR_SUMASLICE: puts("\"sumaslice\" kernel execution failed!!!"); break;
    case ERROR_BACKPROY:  puts("Backprojection kernel execution failed!!!"); break;
  }
  finalizacuda();
  free(proyeccion);
  free(volumen);

  if (ret) exit(ret);
}

int main( int argc, char** argv){
  FileName    fnProyAct,  /// Nombre del archivo de la proyeccion actual 
              fnVolume,   /// Nombre del archivo del volumen reconstruido 
              fnSelFile;  /// Nombre del archivo de seleccion de proyecciones
  bool        initialized;/// Si se ha inicializado el volumen
  Projection  proyAct;    /// Proyeccion actual (leida de un archivo)
  int         Xdim, 
              Ydim, 
              Zdim;       /// Dimensiones del volumen reconstruido
  int x, y, z;
  float *volumen = 0, *proyeccion = 0;

  // Procesa las opciones de linea de comandos
  int hbproy = -1, hbback = -1, numiter = 1;
  float lambda;
  bool sort = false;

  const char* const shortopt = "i:o:l:n:bvs?";
  const struct option longopt[] = {
    { "input",   1, NULL, 'i' },
    { "output",  1, NULL, 'o' },
    { "lambda",  1, NULL, 'l' },
    { "tbproj",  1, NULL, 'P' },
    { "tbback",  1, NULL, 'B' },
    { "blob",    0, NULL, 'b' },
    { "voxel",   0, NULL, 'v' },
    { "noit",    1, NULL, 'n' },
    { "sort",    0, NULL, 's' },
    { "help",    0, NULL, '?' }
  };

  fnVolume = "flipas.vol";
  fnSelFile = "img.sel";
  lambda = 0.01;

  int tipobasis = 0;
  int next_option;
  do{
    next_option = getopt_long(argc, argv, shortopt, longopt, NULL );
    switch( next_option ){
      case 'i': fnSelFile = optarg; break;
      case 'o': fnVolume  = optarg; break;
      case 'l': lambda    = atof(optarg); break;
      case 'P': hbproy    = atoi(optarg); break;
      case 'B': hbback    = atoi(optarg); break;
      case 'b': tipobasis |= 1; break;
      case 'v': tipobasis |= 2; break;
      case 'n': numiter = atoi(optarg); break;
      case 's': sort = true; break;
      case '?': usage(argv[0], 1 ); break;
      case -1 : break;
    }
  }while( next_option != -1 );
  if (optind != argc) usage(argv[0], 1);

  // Parametros por defecto
  if (tipobasis == 0 || tipobasis == 3) {
    puts("You must specify exactly one of -v and -b\n");
    usage(argv[0], 1);
  }
  bool raydriven = tipobasis == 2;




#ifdef borracursor  
  // Oculta el cursor
  initscr();
  curs_set(0);
#endif  


  // Lee el archivo de seleccion
  int numimg = 0;
  vector<string> nombreimg;
  vector<double> dirx, diry, dirz;
  SelFile     sel;        /// Archivo de seleccion de proyecciones 
  try{
    sel.read( fnSelFile );
  } catch (Xmipp_error XE) { acaba(0); cout << XE; exit(1);}
  srand(time(0));

  numimg = sel.ImgNo(); // este nmero cuenta las que no estn activas, pero el next() las da igual
  printf("%i images\n", numimg);
  if (sort) {
    numimg = 0;
    while (!sel.eof()) {
      printf("\rReading angle information... %5i", numimg);
      fflush(stdout);
      string s = sel.NextImg();
      nombreimg.push_back(s);
      ++numimg;

      // Si hay que ordenar, saca los angulos de cada proyeccion
      try { 
        proyAct.read(s, true);
        matrix2D<double> eulerMatrix(3, 3);
        matrix1D<double> direct(3);
        eulerMatrix = proyAct.eulert;
        eulerMatrix.getCol(2, direct);
        dirx.push_back(direct(0));
        diry.push_back(direct(1));
        dirz.push_back(direct(2));
      } catch (Xmipp_error XE) { acaba(0); cout << XE; exit(1);}
    }
  }
  if (sort) printf("\n\rSorting...\n\r");
  fflush(stdout);

  // Ordena
  int indice[numimg], restantes[numimg], quedan = numimg;
  sort &= numimg > 2;
  if (sort) {
    int i, j, k;
    for (i = 0; i < numimg; ++i) restantes[i] = i;
    random_shuffle(&restantes[0], &restantes[numimg]);

    // Elige la primera al azar
    indice[0] = restantes[--quedan];

    /* Elige las restantes para, en cada paso, maximizar el minimo angulo con las dos anteriores
     * (minimizar el maximo valor absoluto del producto escalar) */
    for (i = 1; i < numimg; ++i) {
      int mejor;
      double res = 2;
      for (j = 0; j < quedan; ++j) {
        // Calcula el mayor valor absoluto del producto escalar con las dos anteriores
        double t = -2;
        for (k = max(0, i - 2); k < i; ++k) {
          double presc = dirx[restantes[j]] * dirx[indice[k]] + 
                         diry[restantes[j]] * diry[indice[k]] + 
                         dirz[restantes[j]] * dirz[indice[k]];
          t = max(t, abs(presc));
        }

        if (t < res) {
          mejor = j;
          res = t;
        }
      }

      swap(restantes[mejor], restantes[--quedan]);
      indice[i] = restantes[quedan];
    }
  } else {
    nombreimg.resize(numimg);
    for (int i = 0; i < numimg; ++i) indice[i] = i;
  }

  // Hace el SART
  initialized = false;
  numiter = max(numiter, 1);
  printf("Running %i iteration(s) with lambda = %lf\n", numiter, lambda);
  for (int iteracion = 0; iteracion < numiter; ++iteracion) {
    for (int posimg = 0; posimg < numimg; ++posimg) {
      matrix2D<double> eulerMatrix(3,3); ///Matrix de Euler asociada a la rotacion
      matrix1D<double> direct(3);        ///Vector director de los rayos de la proyeccion

      if (sort)
        fnProyAct = nombreimg[indice[posimg]];
      else {
        nombreimg[posimg] = fnProyAct = sel.NextImg();
        if (posimg == numimg - 1) sort = true; // ya has leido todos los nombres
      }

      //Lee la proyeccion actual y la almacena en 'proyAct'
      try{ 
        bool apply_shifts = true;
        proyAct.read( fnProyAct, apply_shifts );
      }catch (Xmipp_error XE) { acaba(0); cout << XE; exit(1);}

      if( !initialized ){
        proyAct().get_dim( Ydim, Xdim );
        Zdim = max(Xdim, Ydim);
        printf("\rVolume dimensions (z, y, x): %i %i %i\n\r", Zdim, Ydim, Xdim);

        volumen = (float *)malloc(Zdim * Ydim * Xdim * sizeof(*volumen));
        proyeccion = (float *)calloc(Ydim * Xdim * sizeof(*proyeccion), 1);
        inicializacuda(volumen, Zdim, Ydim, Xdim);
        initialized = true;

        if (hbproy < 0) {
          hbproy = 256;        
//          if (Zdim < 81) hbproy = 224; 
  //        else if (Zdim < 192) hbproy = 96;
    //      else hbproy = 288;
        }
        if (hbback < 0) {
          hbback = 64;
//          if (Zdim >= 81 && Zdim < 128) hbback = 112;
  //        else hbback = 16;
        }
        printf("tbproj = %i  tbback = %i\n", hbproy, hbback);
      } 
      printf("\r  Iteration %3i     Image %5i     ", iteracion, posimg);
      cout << fnProyAct;
      fflush(stdout);
      
      eulerMatrix = proyAct.eulert;      // matriz de paso de proyeccion a universal
      int pos = 0;
      for (int i = 0; i < Ydim; ++i) for (int j = 0; j < Xdim; ++j, ++pos)
        proyeccion[pos] = proyAct(i, j);
      matrix3d euler;
      for (int i = 0; i < 3; ++i) for (int j = 0; j < 3; ++j)
        euler.val[i][j] = eulerMatrix(i, j);

      //2) Calcular proyeccion del volumen actual
//      int ret = llamagpu(proyeccion, euler, lambda, hbproy, hbback, raydriven);
  //    if (ret) acaba(ret);

      static int furcia = 0;
//      if (++furcia == 1) break;
    }
    puts("\r");
   }
  acaba(0);

  //Escribe el volumen reconstruido
  VolumeXmipp vol(Zdim, Ydim, Xdim);
  int pos = 0;
  for (z = 0; z < Zdim; ++z) for (y = 0; y < Ydim; ++y) for (x = 0; x < Xdim; ++x, ++pos)
    vol(z, y, x) = 1000.0;
  vol.write( fnVolume );

  return 0;
}

